library ABuffHeroSkill requires Table, ABuff, optional StatusEvents

//*****************************************************************
//*  ABUFF HERO SKILL 1.6
//*
//*  written by: Anitarf
//*  requires: -Table
//*            -ABuff
//*
//*  This library automates the use of aBuffs as passive hero
//*  skills. Once an aBuffType is linked to a hero skill using the
//*  NewABuffHeroSkill function, a buff of that type will be
//*  automatically applied on any hero that learns that skill as
//*  well as any hero that revives with that skill already learned.
//*  If you have the StatusEvents library in your map, then this
//*  library will also detect reincarnation events and correctly
//*  reapply the buff on those events as well. It will also remove
//*  the buff(s) from the hero if he uses a tome of retraining.
//*
//*  To define an aBuff hero skill, use this function:
//*
//*    function NewABuffHeroSkill takes integer abilityId, aBuffType buffType returns nothing
//*
//*  The function only accepts buff types in the STANDARD category.
//*****************************************************************

    private function TomeOfRetraining takes integer itemType returns boolean
        // When a hero uses an item this function is called for that item's type,
        // if it returns true then all aBuff hero skills on that hero are removed.
        // If you have a custom tome of retraining in your map, you need to update this function.
        return itemType=='tret'
    endfunction

// END OF CALIBRATION SECTION    
// ================================================================

    private struct skill
        integer abilityId
        aBuffType buffType
        
        private static Table table
        static method create takes integer abilityId, aBuffType buffType returns skill
            local skill this=skill.allocate()
            set .abilityId=abilityId
            set .buffType=buffType
            debug if skill.table[abilityId]!=0 then
            debug    call BJDebugMsg("NewABuffHeroSkill error: ability rawcode already associated with an ABuffHeroSkill!") 
            debug else
            set skill.table[abilityId]=this
            debug endif
            return this
        endmethod
        static method get takes integer abilityId returns skill
            return skill.table[abilityId]
        endmethod
        static method onInit takes nothing returns nothing
            set skill.table=Table.create()
        endmethod
    endstruct

// ================================================================

    private struct hero
        private unit u
        private skill array skills[5]
        private integer skillCount=0
        
        private static HandleTable table
        private static method create takes unit u returns hero
            local hero this=hero.allocate()
            set .u=u
            set hero.table[u]=this
            return this
        endmethod
        private static method get takes unit u returns hero
            return hero.table[u]
        endmethod
        method onDestroy takes nothing returns nothing
            call hero.table.flush(u)
        endmethod

// ================================================================

        private static method learn takes nothing returns boolean
            local integer i=0
            local hero this
            local skill s = skill.get(GetLearnedSkill())
            if s!=0 then
                set this=hero.get(GetTriggerUnit())
                if this==0 then
                    set this=hero.create(GetTriggerUnit())
                endif
                loop
                    exitwhen i>=.skillCount or s==.skills[i]
                    set i=i+1
                endloop
                if i>=.skillCount then
                    set .skills[i]=s
                    set .skillCount=i+1
                endif
                call ABuffApply(s.buffType, .u, .u, 0.0, GetUnitAbilityLevel(.u, s.abilityId), 0)
            endif
            return false
        endmethod

        private static method ressurect takes unit u returns nothing
            local integer i=0
            local hero this=hero.get(u)
            if this!=0 then
                loop
                    exitwhen i>=.skillCount
                    call ABuffApply(skills[i].buffType, .u, .u, 0.0, GetUnitAbilityLevel(.u, skills[i].abilityId), 0)
                    set i=i+1
                endloop
            endif
        endmethod
        private static method revive takes nothing returns boolean
            call hero.ressurect(GetTriggerUnit())
            return false
        endmethod

        private static method retrain takes nothing returns boolean
            local integer i=0
            local hero this=hero.get(GetTriggerUnit())
            if this!=0 and TomeOfRetraining(GetItemTypeId(GetManipulatedItem())) then
                loop
                    exitwhen i>=.skillCount
                    call ABuffRemove(GetABuffFromUnitByType(.u, skills[i].buffType))
                    set i=i+1
                endloop
                call .destroy()
            endif
            return false
        endmethod

        private static method onInit takes nothing returns nothing
            local trigger t
            set t = CreateTrigger(  )
            call TriggerRegisterAnyUnitEventBJ( t, EVENT_PLAYER_HERO_SKILL )
            call TriggerAddCondition( t, Condition(function hero.learn) )
          static if LIBRARY_StatusEvents then
            call OnUnitResurrect(hero.ressurect)
          else
            set t = CreateTrigger(  )
            call TriggerRegisterAnyUnitEventBJ( t, EVENT_PLAYER_HERO_REVIVE_FINISH )
            call TriggerAddCondition( t, Condition(function hero.revive) )
          endif
            set t = CreateTrigger(  )
            call TriggerRegisterAnyUnitEventBJ( t, EVENT_PLAYER_UNIT_USE_ITEM )
            call TriggerAddCondition( t, Condition(function hero.retrain) )

            set hero.table=HandleTable.create()
        endmethod
    endstruct

// ================================================================
//  MAIN USER FUNCTION

    function NewABuffHeroSkill takes integer abilityId, aBuffType buffType returns nothing
        if buffType.category==ABuff_STANDARD then
            call skill.create(abilityId, buffType)
        debug else
            debug call BJDebugMsg("NewABuffHeroSkill error: the function only accepts buff types in the STANDARD category.")
        endif
    endfunction
endlibrary